@@ -1,9 +1,9 @@
 import sqlalchemy
-from sqlalchemy import inspect
+from sqlalchemy import inspect, text
 from sqlalchemy.ext.declarative import base
 from sqlalchemy.engine.reflection import Inspector
 from sqlalchemy.ext.hybrid import hybrid_property
-from sqlalchemy.orm import object_mapper, session, relationships
+from sqlalchemy.orm import object_mapper, session, relationships, RelationshipProperty
 
 import logic_bank
 from logic_bank.rule_bank.rule_bank import RuleBank
@@ -13,7 +13,10 @@ from sqlalchemy.ext.declarative import declarative_base
 from logic_bank.rule_bank import rule_bank_withdraw
 from logic_bank.rule_type.constraint import Constraint
 from logic_bank.rule_type.formula import Formula
+from logic_bank.rule_type.parent_cascade import ParentCascade, ParentCascadeAction
+from logic_bank.rule_type.parent_check import ParentCheck
 from logic_bank.rule_type.row_event import EarlyRowEvent
+from logic_bank.util import ConstraintException
 
 
 class LogicRow:
@@ -54,8 +57,7 @@ class LogicRow:
 
         rb = RuleBank()
         self.rb = rb
-        self.session = rb._session
-        self.engine = rb._engine
+        self.session = a_session
         self.some_base = declarative_base()
 
         self.name = type(self.row).__name__  # class name (not table name)
@@ -66,8 +68,6 @@ class LogicRow:
             else:
                 self.table_meta = inspect(self.row)
                 self.log_engine("Restriction: Class Name must equal Table Name: " + self.name)
-        if self.engine is not None:  # e.g, for testing legacy logic (no RuleBank)
-            self.inspector = Inspector.from_engine(self.engine)
 
     def get_attr_name(self, mapper, attr)-> str:
         """polymorhpism is for wimps - find the name
@@ -144,13 +144,14 @@ class LogicRow:
         result += f'  row@: {str(hex(id(self.row)))}'
         return result  # str(my_dict)
 
-    def log(self, msg: str):
+    def log(self, msg: str) -> str:
         """
         prints to logic_logger: row/old_row, indented, inserting msg
         """
         output = str(self)
         output = output.replace("]:", "] {" + msg + "}", 1)
         logic_bank.logic_logger.debug(output)  # more on this later
+        return output
 
     def log_engine(self, msg: str):
         """
@@ -170,7 +171,7 @@ class LogicRow:
         result = None
         if a_row is not None:
             result_class = a_row.__class__
-            result = result_class()
+            result = result_class()  # TODO - this requires empty CTOR... add factory?
             row_mapper = object_mapper(a_row)
             for each_attr in row_mapper.all_orm_descriptors:
                 each_attr_name = self.get_attr_name(mapper=row_mapper, attr=each_attr)
@@ -207,7 +208,7 @@ class LogicRow:
 
     def early_row_events(self):
         self.log_engine("early_events")
-        early_row_events = rule_bank_withdraw.generic_rules_of_class(EarlyRowEvent)
+        early_row_events = rule_bank_withdraw.generic_rules_of_class(EarlyRowEvent)  # FIXME - review duplication
         for each_row_event in early_row_events:
             each_row_event.execute(self)
         early_row_events = rule_bank_withdraw.rules_of_class(self, EarlyRowEvent)
@@ -270,16 +271,20 @@ class LogicRow:
 
     def cascade_delete_children(self):
         """
-        Find child relationships that are cascade delete, and delete the children.
+        This recursive descent is required to adjust dependent sums/counts on passive_deletes; ie,
 
-        This recursive descent is required to adjust dependent sums/counts.
+        when (and only when) the DBMS - and *not* SQLAlchemy - does the deletes.
+
+        (When SQLAlchemy does deletes, these are queued through the normal delete logic.)
+        @see nw/tests/test_dlt_order.py
         """
+
         parent_mapper = object_mapper(self.row)
         my_relationships = parent_mapper.relationships
         for each_relationship in my_relationships:  # eg, cust has child OrderDetail
             if each_relationship.direction == sqlalchemy.orm.interfaces.ONETOMANY:  # eg, OrderDetail
                 child_role_name = each_relationship.key  # eg, OrderList
-                if each_relationship.cascade.delete:
+                if each_relationship.cascade.delete and each_relationship.passive_deletes:
                     child_rows = getattr(self.row, child_role_name)
                     for each_child_row in child_rows:
                         old_child = self.make_copy(each_child_row)
@@ -289,9 +294,106 @@ class LogicRow:
                                                         nest_level=1 + self.nest_level,
                                                         a_session=self.session,
                                                         row_sets=self.row_sets)
-                        each_child_logic_row.delete(reason="Cascade Delete - " + child_role_name)
+                        each_child_logic_row.delete(reason="Cascade Delete to run rules on - " + child_role_name)
+                        self.session.delete(each_child_row)  # deletes in beforeFlush are not re-queued
+        enforce_cascade = False
+        if enforce_cascade:  # disabled - SQLAlchemy DOES enforce cascade delete/nullify; prevent way less important
+            """
+            per parent_cascade rule(s), nullify (child FKs), delete (children), prevent (if children exist)
 
-    def cascade_to_children(self):
+            Default is ParentCascadeAction.PREVENT.
+
+            This recursive descent is required to adjust dependent sums/counts.
+            """
+            list_parent_cascade_rules = rule_bank_withdraw.rules_of_class(self, ParentCascade)
+            defined_relns = {}
+            for each_parent_cascade_rule in list_parent_cascade_rules:
+                defined_relns[each_parent_cascade_rule._relationship] = each_parent_cascade_rule
+            for each_relationship in my_relationships:  # eg, Order has child OrderDetail
+                if each_relationship.direction == sqlalchemy.orm.interfaces.ONETOMANY:  # eg, OrderDetail
+                    each_child_role_name = each_relationship.key  # eg, OrderDetailList
+                    refinteg_action = ParentCascadeAction.PREVENT
+                    if each_child_role_name in defined_relns:
+                        refinteg_action = defined_relns[each_child_role_name]._action
+                    child_rows = getattr(self.row, each_child_role_name)
+                    for each_child_row in child_rows:
+                        old_child = self.make_copy(each_child_row)
+                        each_child_logic_row = LogicRow(row=each_child_row,
+                                                        old_row=old_child,
+                                                        ins_upd_dlt="dlt",
+                                                        nest_level=1 + self.nest_level,
+                                                        a_session=self.session,
+                                                        row_sets=self.row_sets)
+
+                        if refinteg_action == ParentCascadeAction.DELETE:  # each_relationship.cascade.delete:
+                            each_child_logic_row.delete(reason="Cascade Delete - " + each_child_role_name)
+
+                        elif refinteg_action == ParentCascadeAction.NULLIFY:
+                            for p, c in each_relationship.local_remote_pairs:
+                                setattr(each_child_row, c.name, None)
+                            each_child_logic_row.update(reason="Cascade Nullify - " + each_child_role_name)
+
+                        elif refinteg_action == ParentCascadeAction.PREVENT:
+                            raise ConstraintException("Delete rejected - " + each_child_role_name + " has rows")
+                        else:
+                            raise Exception("Invalid parent_cascade action: " + refinteg_action)
+
+    def is_primary_key_changed(self) -> bool:
+        meta = self.table_meta
+        if hasattr(meta, "primary_key"):
+            pk_columns = meta.primary_key.columns
+        else:
+            pk_columns = meta.mapper.mapped_table.primary_key.columns
+        if len(pk_columns) == 0:
+            raise Exception("No Primary Key: " + self.__str__())
+        for each_pk_column in pk_columns:
+            each_child_column_name = each_pk_column.name
+            if getattr(self.row, each_child_column_name) != getattr(self.old_row, each_child_column_name):
+                return True
+        return False
+
+    def get_old_child_rows(self, relationship: RelationshipProperty):
+        """
+        result = getattr(self.old_row, role_name)  # even with util.use_transient, yields nothing, unsure why
+        """
+
+        child_filter = {}
+        for p, c in relationship.local_remote_pairs:
+            child_filter[c.name] = getattr(self.old_row, p.name)
+        result = self.session.query(relationship.mapper.entity).filter_by(**child_filter).all()
+        return result
+
+    def parent_cascade_pk_change(self):  # ???
+        """
+        cascade pk change (if any) to children, unconditionally.
+
+        Presumption: children ref the same pKey (vs. some other "candidate key")
+        """
+        if self.is_primary_key_changed():
+            list_parent_cascade_rules = rule_bank_withdraw.rules_of_class(self, ParentCascade)
+            defined_relns = {}
+            for each_parent_cascade_rule in list_parent_cascade_rules:
+                defined_relns[each_parent_cascade_rule._relationship] = each_parent_cascade_rule
+            parent_mapper = object_mapper(self.row)
+            my_relationships = parent_mapper.relationships
+            for each_relationship in my_relationships:  # eg, order has parents cust & emp, child orderdetail
+                if each_relationship.direction == sqlalchemy.orm.interfaces.ONETOMANY:  # cust, emp
+                    reason = "Cascading PK change to: " +\
+                             each_relationship.backref + "->" +\
+                             each_relationship.key
+                    child_rows = self.get_old_child_rows(relationship = each_relationship)
+                    for each_child_row in child_rows:
+                        old_child = self.make_copy(each_child_row)
+                        each_child_logic_row = LogicRow(row=each_child_row, old_row=old_child, ins_upd_dlt="upd",
+                                                        nest_level=1 + self.nest_level,
+                                                        a_session=self.session, row_sets=self.row_sets)
+                        for p, c in each_relationship.local_remote_pairs:
+                            setattr(each_child_row, c.name, getattr(self.row, p.name))
+                        each_child_logic_row.update(reason=reason)
+
+        return self
+
+    def parent_cascade_attribute_changes_to_children(self):
         """
         Child Formulas can reference (my) Parent Attributes, so...
         If the *referenced* Parent Attributes are changed, cascade to child
@@ -404,36 +506,84 @@ class LogicRow:
         for each_constraint in constraint_rules:
             each_constraint.execute(self)
 
-    def load_parents(self):
-        """ sqlalchemy lazy does not work for inserts... do it here...
+    def is_foreign_key_null(self, relationship: sqlalchemy.orm.relationships) -> bool:
+        child_columns = relationship.local_columns
+        if len(child_columns) == 0:
+            raise Exception("Malformed relationship has no foreign key: " +
+                            str(relationship))
+        for each_child_column in child_columns:
+            each_child_column_name = each_child_column.name
+            if getattr(self.row, each_child_column_name) is None:
+                return True
+        return False
+
+    def load_parents_on_insert(self):
+        """ sqlalchemy lazy does not work for inserts... do it here because...
         1. RI would require the sql anyway
         2. Provide a consistent model - your parents are always there for you
             - eg, see add_order event rule - references {sales_rep.Manager.FirstName}
         """
-        def is_foreign_key_null(relationship: sqlalchemy.orm.relationships):
-            child_columns = relationship.local_columns
-            if len(child_columns) == 0:
-                raise Exception("Malformed relationship has no foreign key: " +
-                                str(relationship))
-            for each_child_column in child_columns:
-                each_child_column_name = each_child_column.name
-                if getattr(self.row, each_child_column_name) is None:
-                    return True
-            return False
+
+        ref_integ_enabled = True
+        list_ref_integ_rules = rule_bank_withdraw.rules_of_class(self, ParentCheck)
+        if list_ref_integ_rules:
+            ref_integ_rule = list_ref_integ_rules[0]
 
         child_mapper = object_mapper(self.row)
         my_relationships = child_mapper.relationships
         for each_relationship in my_relationships:  # eg, order has parents cust & emp, child orderdetail
             if each_relationship.direction == sqlalchemy.orm.interfaces.MANYTOONE:  # cust, emp
                 parent_role_name = each_relationship.key  # eg, OrderList
-                if is_foreign_key_null(each_relationship) is False:
+                if self.is_foreign_key_null(each_relationship) is False:
                     # continue
                     self.get_parent_logic_row(parent_role_name)  # sets the accessor
                     does_parent_exist = getattr(self.row, parent_role_name)
-                    if not does_parent_exist:
+                    if does_parent_exist is None and ref_integ_enabled:
                         msg = "Missing Parent: " + parent_role_name
                         self.log(msg)
-                        raise Exception(msg)
+                        raise ConstraintException(msg)
+                    else:
+                        self.log("Warning: Missing Parent: " + parent_role_name)
+                        pass # if you don't care, I don't care
+        return self
+
+    def check_parents_on_update(self):
+        """ per ParentCheck rule, verify parents exist.
+
+        If disabled, ignore (with warning).
+        """
+
+        list_ref_integ_rules = rule_bank_withdraw.rules_of_class(self, ParentCheck)
+        if list_ref_integ_rules:
+            ref_integ_rule = list_ref_integ_rules[0]
+            if ref_integ_rule._enable:
+                child_mapper = object_mapper(self.row)
+                my_relationships = child_mapper.relationships
+                for each_relationship in my_relationships:  # eg, order has parents cust & emp, child orderdetail
+                    if each_relationship.direction == sqlalchemy.orm.interfaces.MANYTOONE:  # cust, emp
+                        parent_role_name = each_relationship.key  # eg, OrderList
+                        if not self.is_foreign_key_null(each_relationship):
+                            # continue
+                            reason = "Cascading PK change to: " + \
+                                     each_relationship.key + "->" + \
+                                     each_relationship.back_populates
+                            if self.reason == reason:
+                                """
+                                The parent doing the cascade obviously exists,
+                                and note: try to getattr it will fail
+                                (FIXME design review - perhaps SQLAlchemy is not checking cache?)
+                                """
+                                pass
+                            else:
+                                self.get_parent_logic_row(parent_role_name)  # sets the accessor
+                                does_parent_exist = getattr(self.row, parent_role_name)
+                                if does_parent_exist is None and ref_integ_rule._enable == True:
+                                    msg = "Missing Parent: " + parent_role_name
+                                    self.log(msg)
+                                    raise ConstraintException(msg)
+                                else:
+                                    self.log("Warning: Missing Parent: " + parent_role_name)
+                                    pass # if you don't care, I don't care
         return self
 
     def adjust_parent_aggregates(self):
@@ -455,7 +605,7 @@ class LogicRow:
                                                  parent_role_name=each_parent_role)
             for each_aggregate in each_aggr_list:
                 each_aggregate.adjust_parent(parent_adjuster)  # adjusts each_parent iff req'd
-            parent_adjuster.save_altered_parents()
+            parent_adjuster.save_altered_parents()  # iff req'd (altered only)
 
     def user_row_update(self, row: base, ins_upd_dlt: str) -> 'LogicRow':
         result_logic_row = LogicRow(row = row,
@@ -480,11 +630,13 @@ class LogicRow:
             self.reason = reason
             self.log("Update - " + reason)
             self.early_row_events()
+            self.check_parents_on_update()
             self.copy_rules()
             self.formula_rules()
             self.adjust_parent_aggregates()  # parent chaining (sum / count adjustments)
             self.constraints()
-            self.cascade_to_children()  # child chaining (cascade changed parent references)
+            self.parent_cascade_attribute_changes_to_children()  # child chaining (cascade changed parent references)
+            self.parent_cascade_pk_change()  # actions - delete, nullify, prevent
             if self.row_sets is not None:  # eg, for debug as in upd_order_shipped test
                 self.row_sets.remove_submitted(logic_row=self)
 
@@ -501,7 +653,7 @@ class LogicRow:
         else:
             self.reason = reason
             self.log("Insert - " + reason)
-            self.load_parents()
+            self.load_parents_on_insert()
             self.early_row_events()
             self.copy_rules()
             self.formula_rules()
@@ -530,7 +682,7 @@ class LogicRow:
 
 class ParentRoleAdjuster:
     """
-    Contains current / previous parent_logic_row
+    Contains child_logic_row and current / previous parent_logic_row
         Set iff parent needs adjustment
     and method to save_altered_parents.
 
