@@ -4,11 +4,13 @@ from sqlalchemy.orm.attributes import InstrumentedAttribute
 from sqlalchemy.orm import session
 
 from logic_bank.rule_bank import rule_bank_withdraw  # reduce circular imports
-from logic_bank.rule_bank.rule_bank_setup import setup, validate
+import logic_bank.rule_bank.rule_bank_setup as rule_bank_setup
 from logic_bank.rule_type.constraint import Constraint
 from logic_bank.rule_type.copy import Copy
 from logic_bank.rule_type.count import Count
 from logic_bank.rule_type.formula import Formula
+from logic_bank.rule_type.parent_cascade import ParentCascade, ParentCascadeAction
+from logic_bank.rule_type.parent_check import ParentCheck
 from logic_bank.rule_type.row_event import EarlyRowEvent, RowEvent, CommitRowEvent
 from logic_bank.rule_type.sum import Sum
 
@@ -17,24 +19,28 @@ class LogicBank:
     """
     Logic consists of Rules, and Python.
 
-    Activate your logic,
-    providing a function that declares your rules and Python.
-    """
+    Activate your logic by calling
 
+        activate(session: session, activator: my_logic)
+
+    where myLogic is a function that declares your rules and Python.
+    """
+    @staticmethod
     def activate(session: session, activator: callable):
         """
-        load rules - executed on commit
+        register SQLAlchemy listeners
+
+        create RuleBank, load rules - later executed on commit
 
         raises exception if cycles detected
 
         :param session: SQLAlchemy session
-        :param activator: function that declares rules (e.g., Rule.sum...)
-        :return:
+        :param activator: user function that declares rules (e.g., Rule.sum...)
         """
-        engine = session.bind.engine
-        setup(session, engine)
+
+        rule_bank_setup.setup(session)
         activator()
-        validate(session, engine)
+        rule_bank_setup.compute_formula_execution_order()
 
 
 class Rule:
@@ -80,8 +86,64 @@ class Rule:
         Example
           Rule.constraint(validate=Customer, as_condition=lambda row: row.Balance <= row.CreditLimit,
                           error_msg="balance ({row.Balance}) exceeds credit ({row.CreditLimit})")
+
+
+        Constraint failures raise ConstraintException, e.g.:
+            try:
+                session.commit()
+            except ConstraintException as ce:
+                print("Constraint raised: " + str(ce))
+
         """
-        return Constraint(validate=validate, calling=calling, as_condition=as_condition, error_msg=error_msg)  # --> load_logic
+        return Constraint(validate=validate, calling=calling, as_condition=as_condition, error_msg=error_msg)
+
+    @staticmethod
+    def parent_check(validate: object,
+                     error_msg: str = "(error_msg not provided)",
+                     enable: bool = True):
+        """
+        Parent Checks ensure that non-null foreign keys are present in parent class
+
+        Example
+           Rule.parent_check(validate=Customer, enable=True, error_msg="Missing Parent")
+
+        Use enable: False to tolerate orphans
+            Not recommended - for existing databases with bad data
+            Behavior is undefined for other rules (sum, count, parent references, etc)
+
+        Parent_check failures raise ConstraintException, e.g.:
+            try:
+                session.commit()
+            except ConstraintException as ce:
+                print("Constraint raised: " + str(ce))
+
+        """
+        return ParentCheck(validate=validate, error_msg=error_msg, enable=enable)
+
+    '''
+    disabled, per ORM support (retained in case of misunderstandings)
+        @staticmethod
+        def parent_cascade(validate: object,
+                           error_msg: str = "(error_msg not provided)",
+                           relationship: str = "*",
+                           action: ParentCascadeAction = ParentCascadeAction.NULLIFY):
+            """
+            Parent Cascade specifies processing for child rows on parent delete
+    
+            Example
+               Rule.parent_cascade(validate=Order, relationship="OrderDetailList", action=ParentCascadeAction.DELETE)
+    
+            If rule or action not specified, default is ParentCascadeAction.NULLIFY
+    
+            Parent_cascade with ParentCascadeAction.NULLIFY can raise ConstraintException, e.g.:
+                try:
+                    session.commit()
+                except ConstraintException as ce:
+                    print("Constraint raised: " + str(ce))
+    
+            """
+            return ParentCascade(validate=validate, error_msg=error_msg, relationship=relationship, action=action)
+    '''
 
     @staticmethod
     def formula(derive: InstrumentedAttribute, calling: Callable = None,
